==> file4life-1.0.2-source.tgz: file4life-1.0.2/src/F4L/Knowledge.hs <==
{- This file is part of file4life, a file backup program written in Haskell.
 - file4life is Copyright (C) Antonio Bellezza 2014
 - file4life is free software: you can redistribute it and/or modify
 - it under the terms of the GNU General Public License as published by
 - the Free Software Foundation, either version 3 of the License, or
 - (at your option) any later version.

 - file4life is distributed in the hope that it will be useful,
 - but WITHOUT ANY WARRANTY; without even the implied warranty of
 - MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 - GNU General Public License for more details.

 - You should have received a copy of the GNU General Public License
 - along with file4life.  If not, see <http://www.gnu.org/licenses/>.
 -}
{-# LANGUAGE GeneralizedNewtypeDeriving, FlexibleInstances, OverlappingInstances #-}

module F4L.Knowledge where

import F4L
import F4L.SeenTree as ST
import F4L.Op
import F4L.Utils

import Control.DeepSeq
import Control.Lens
import Control.Monad
import qualified Data.Binary as B
-- import qualified Data.ByteString.Lazy as BL
-- import qualified Data.ByteString.Short as SBS
import Data.List( find )
import qualified Data.Map.Strict as SM
import Data.Maybe( mapMaybe )
import Data.Set( Set )
import qualified Data.Set as S
-- import qualified Data.Text as T


data FileData = FileData { fd_seen :: !( Set ( StorageId, ShortPath ) )
                         , fd_priority :: !Priority
                         }
                deriving Show


noFileData = FileData S.empty Skip


type StorageSeenMap = SM.Map StorageId [ ( EnvironmentId, FilePath ) ]


storagesSeenIn envid = SM.map ( map snd . filter ( ( envid == ) . fst ) )


instance B.Binary FileData where
  put ( FileData a b ) = B.put a >> B.put b
  get = liftM2 FileData B.get B.get

instance NFData FileData where
  rnf ( FileData a b ) = rnf a `seq` rnf b

class Knowledge k where
  noKnowledge :: k
  applyFact :: Fact -> k -> k
  applyFact = const id
  kCode :: k -> String

applyFact' f s = f `deepseq` applyFact f s

instance Knowledge () where
  noKnowledge = ()
  kCode = const ""

instance ( Knowledge a, Knowledge b ) => Knowledge ( a, b ) where
  noKnowledge = ( noKnowledge, noKnowledge )
  applyFact f ( a, b ) = -- TODO: more concise way to force strictness?
    let 
      na = applyFact f a
      nb = applyFact f b
    in
     na `seq` nb `seq` ( na, nb )
  kCode ( a, b ) = "C" ++ kCode a ++ "_" ++ kCode b ++ "D"


class ( Knowledge k, B.Binary k ) => SerializableKnowledge k where
  putK :: FidStore -> k -> B.Put
  putK = const B.put

  getK :: FidStore -> B.Get k
  getK = const B.get


instance ( SerializableKnowledge k1, SerializableKnowledge k2 ) => SerializableKnowledge (k1, k2) where
  putK f (k1, k2) = putK f k1 >> putK f k2
  getK f = do
    k1 <- getK f
    k2 <- getK f
    return (k1, k2)

---------------------------------------- KnownData
-- TODO: change name
data KnownData = KnownData { 
    _kd_files :: !(SM.Map FileId FileData)
  , _kd_storages :: !StorageSeenMap
  , _kd_contents :: !(SM.Map StorageId SeenTree)
  } deriving Show

class KnownDataClass a where
  kd_files :: Lens' a (SM.Map FileId FileData)
  kd_storages :: Lens' a StorageSeenMap
  kd_contents :: Lens' a (SM.Map StorageId SeenTree )

instance KnownDataClass KnownData where
  kd_files = lens _kd_files $ \ s a -> s { _kd_files = a }
  kd_storages = lens _kd_storages $ \ s a -> s { _kd_storages = a }
  kd_contents = lens _kd_contents $ \ s a -> s { _kd_contents = a }

instance KnownDataClass (KnownData, b) where
  kd_files = _1 . kd_files
  kd_storages = _1 . kd_storages
  kd_contents = _1 . kd_contents

instance KnownDataClass a => KnownDataClass (b, a) where
  kd_files = _2 . kd_files
  kd_storages = _2 . kd_storages
  kd_contents = _2 . kd_contents

instance Knowledge KnownData where
  noKnowledge = KnownData SM.empty SM.empty SM.empty
  applyFact = applyFact''
  kCode = const "k"

instance SerializableKnowledge KnownData where
  putK f ( KnownData a b c ) = do
    B.put b
    B.put $ map ( mapFst $ fromJust' "Shorthand" . flip findShorthand f ) $ SM.toList a
    B.put $ toShared f c
    where
      toShared f = map ( mapSnd ( toShared' f ) ) . SM.toList 
      toShared' f = fmap ( fromJust' "Shorthand" . flip findShorthand f ) . toLazy
  getK f = do
    b <- getMap getStrict getStrict
    let sids = S.toList $ S.fromList $ SM.keys b
    let getSid sid = fromJust' "Deduplicate sid" $ find ( == sid ) sids
    let dedup fd = fd { fd_seen = S.fromList $ map ( mapFst getSid ) $ S.toList $ fd_seen fd }
    a <- getMap ( liftM ( force . head . flip findFid f ) B.get ) ( liftM dedup $ getStrict )
    c <- getMap getStrict $ ST.getSeenTree ( head . flip findFid f )
    return $ KnownData a b c


applyFact'' :: Fact -> KnownData -> KnownData
applyFact'' ( FileSeen _ fid sid path ) = ( kd_contents %~ addToContent ) . ( kd_files %~ f )
  where
    f fd = SM.insert fid ( pre { fd_seen = S.insert ( sid, pathToShort path ) $ fd_seen pre } ) fd
      where
        pre = SM.findWithDefault noFileData fid fd
    addToContent = SM.insertWith ( const $ insertFile path fid ) sid ( insertFile path fid ST.empty )
applyFact'' ( FileMiss _ sid path ) = ( kd_contents %~ remFromContent )
  where
    -- TODO: also delete from kd_files
    remFromContent = SM.adjust ( removeFile path ) sid
applyFact'' ( FileAttr _ fid attrs ) = kd_files %~ f
  where
    f fd =
      case mapMaybe ( \ a -> case a of Require req -> Just req; _ -> Nothing ) attrs of
        [] -> fd
        req : _ ->
          SM.insert fid ( pre { fd_priority = updatePriority req ( fd_priority pre ) } ) fd
      where
        pre = SM.findWithDefault noFileData fid fd
applyFact'' ( StorageSeen _ts envid sid path ) = kd_storages %~ storageInsert envid sid path
  where
    storageInsert :: EnvironmentId -> StorageId -> FilePath -> StorageSeenMap -> StorageSeenMap
    storageInsert envid sid fp m =
      flip ( SM.insert sid ) m $ case SM.lookup sid m of
        Nothing -> [ (envid, fp) ]
        Just l -> ( envid, fp ) : filter ( ( envid, fp ) /= ) l
applyFact'' _ = id

instance B.Binary KnownData where
  put ( KnownData a b c ) = B.put a >> B.put b >> B.put c
  get = liftM3 KnownData B.get B.get B.get


---------------------------------------- FactCounter
data FactCounter = FactCounter { fc_c :: !Int }
                  deriving (Show, Eq)

class FactCounterClass a where
  fact_counter :: Lens' a Int

instance FactCounterClass FactCounter where
  fact_counter = lens fc_c $ const FactCounter

instance FactCounterClass ( FactCounter, b ) where
  fact_counter = _1 . fact_counter

instance FactCounterClass a => FactCounterClass ( b, a ) where
  fact_counter = _2 . fact_counter

instance Knowledge FactCounter where
  noKnowledge = FactCounter 0
  applyFact _ ( FactCounter i ) = FactCounter ( i + 1 )
  kCode = const "c"

instance B.Binary FactCounter where
  put = B.put . fc_c
  get = liftM FactCounter B.get

instance SerializableKnowledge FactCounter


---------------------------------------- FileSet
data FileSet = FileSet {
  _fs_files :: !( S.Set FileId ) 
  } deriving Show

class FileSetClass a where
  fs_files :: Lens' a ( S.Set FileId )

instance FileSetClass FileSet where
  fs_files = lens _fs_files $ const FileSet

instance FileSetClass ( FileSet, b ) where
  fs_files = _1 . fs_files

instance FileSetClass a => FileSetClass ( b, a ) where
  fs_files = _2 . fs_files

instance Knowledge FileSet where
  noKnowledge = FileSet S.empty
  applyFact f x @ ( FileSet s ) = case mfact_file f of
    Just fid -> FileSet $ S.insert fid s
    Nothing -> x
  kCode = const "f"

instance B.Binary FileSet where
  put = B.put . _fs_files
  get = liftM FileSet B.get

instance SerializableKnowledge FileSet


---------------------------------------- SeenStorages
data SeenStorages = SeenStorages {
  _ss_storages :: !( SM.Map StorageId [ String ] )
  } deriving Show

class SeenStoragesClass a where
  seen_storages :: Lens' a ( SM.Map StorageId [ String ] )

instance SeenStoragesClass SeenStorages where
  seen_storages = lens _ss_storages $ \ s a -> s { _ss_storages = a }

instance SeenStoragesClass ( SeenStorages, b ) where
  seen_storages = _1 . seen_storages

instance SeenStoragesClass a => SeenStoragesClass ( b, a ) where
  seen_storages = _2 . seen_storages

instance Knowledge SeenStorages where
  noKnowledge = SeenStorages SM.empty
  applyFact ( LabelStorage sid label ) ( SeenStorages s ) =
    SeenStorages ( SM.insertWith (++) sid [ label ] s )
  applyFact x ( SeenStorages s ) =
    case mfact_storage x of
      Just sid -> sid `deepseq` SeenStorages ( SM.insertWith (++) sid [] s )
      Nothing -> SeenStorages s
  kCode = const "s"

instance B.Binary SeenStorages where
  put = B.put . _ss_storages
  get = liftM SeenStorages B.get

instance SerializableKnowledge SeenStorages


---------------------------------------- SeenEnvironments
data SeenEnvironments = SeenEnvironments {
  _se_environments :: !( SM.Map StorageId [ String ] )
  } deriving Show

class SeenEnvironmentsClass a where
  seen_environments :: Lens' a ( SM.Map StorageId [ String ] )

instance SeenEnvironmentsClass SeenEnvironments where
  seen_environments = lens _se_environments $ \ s a -> s { _se_environments = a }

instance SeenEnvironmentsClass ( SeenEnvironments, b ) where
  seen_environments = _1 . seen_environments

instance SeenEnvironmentsClass a => SeenEnvironmentsClass ( b, a ) where
  seen_environments = _2 . seen_environments

instance Knowledge SeenEnvironments where
  noKnowledge = SeenEnvironments SM.empty
  applyFact ( LabelEnvironment envid label ) ( SeenEnvironments s ) =
    SeenEnvironments ( SM.insertWith (++) envid [ label ] s )
  applyFact x @ ( StorageSeen _ _ _ _ ) ( SeenEnvironments s ) =
    SeenEnvironments ( SM.insertWith (++) ( fact_env x ) [] s )
  applyFact _ s = s
  kCode = const "e"

instance B.Binary SeenEnvironments where
  put = B.put . _se_environments
  get = liftM SeenEnvironments B.get

instance SerializableKnowledge SeenEnvironments

==> file4life-1.1.0-source.tgz: file4life-1.1.0/src/F4L/Knowledge.hs <==
{- This file is part of file4life, a file backup program written in Haskell.
 - file4life is Copyright (C) Antonio Bellezza 2014
 - file4life is free software: you can redistribute it and/or modify
 - it under the terms of the GNU General Public License as published by
 - the Free Software Foundation, either version 3 of the License, or
 - (at your option) any later version.

 - file4life is distributed in the hope that it will be useful,
 - but WITHOUT ANY WARRANTY; without even the implied warranty of
 - MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 - GNU General Public License for more details.

 - You should have received a copy of the GNU General Public License
 - along with file4life.  If not, see <http://www.gnu.org/licenses/>.
 -}
{-# LANGUAGE GeneralizedNewtypeDeriving, FlexibleInstances, OverlappingInstances #-}

module F4L.Knowledge where

import F4L
import F4L.Op
import F4L.PathAttributes as PA
import F4L.SeenTree as ST
import F4L.Utils

import Control.DeepSeq
import Control.Lens
import Control.Monad
import Control.Monad.State
import qualified Data.Binary as B
-- import qualified Data.ByteString.Lazy as BL
-- import qualified Data.ByteString.Short as SBS
import Data.List( find )
import qualified Data.Map.Strict as SM
import Data.Maybe( fromMaybe, mapMaybe )
import Data.Set( Set )
import qualified Data.Set as S
-- import qualified Data.Text as T
import System.FilePath( joinPath, splitDirectories )


data FileData = FileData { fd_seen :: !( Set ( StorageId, ShortPath ) )
                         , fd_priorities :: !( SM.Map Priority Int )
                         }
                deriving Show


fd_priority fd = foldr max Skip $ map fst $ filter ( ( > 0 ) . snd ) $ SM.toList $ fd_priorities fd

fd_withPriority :: Priority -> FileData -> FileData
fd_withPriority pri fd = fd { fd_priorities = SM.insertWith (+) pri 1 $ fd_priorities fd }

fd_swapPriority :: Priority -> Priority -> FileData -> FileData
fd_swapPriority pre post fd =
  fd_withPriority post $ fd { fd_priorities = SM.alter dec pre ( fd_priorities fd ) }
  where
    dec Nothing = error $ "Trying to replace unset priority " ++ show pre ++ " for " ++ show fd
    dec ( Just i ) | i <= 0 = error $ "Wrong multiplicity for priority " ++ show pre ++ " in " ++ show fd 
    dec ( Just 1 ) = Nothing
    dec ( Just i ) = Just $! i - 1


noFileData = FileData S.empty SM.empty


type StorageSeenMap = SM.Map StorageId [ ( EnvironmentId, FilePath ) ]


storagesSeenIn envid = SM.map ( map snd . filter ( ( envid == ) . fst ) )


instance B.Binary FileData where
  put ( FileData a b ) = B.put a >> B.put b
  get = liftM2 FileData B.get B.get

instance NFData FileData where
  rnf ( FileData a b ) = rnf a `seq` rnf b

class Knowledge k where
  noKnowledge :: k
  applyFact :: Fact -> k -> k
  applyFact = const id
  kCode :: k -> String

applyFact' f s = f `deepseq` applyFact f s

instance Knowledge () where
  noKnowledge = ()
  kCode = const ""

instance ( Knowledge a, Knowledge b ) => Knowledge ( a, b ) where
  noKnowledge = ( noKnowledge, noKnowledge )
  applyFact f ( a, b ) = -- TODO: more concise way to force strictness?
    let 
      na = applyFact f a
      nb = applyFact f b
    in
     na `seq` nb `seq` ( na, nb )
  kCode ( a, b ) = "C" ++ kCode a ++ "_" ++ kCode b ++ "D"


class ( Knowledge k, B.Binary k ) => SerializableKnowledge k where
  putK :: FidStore -> k -> B.Put
  putK = const B.put

  getK :: FidStore -> B.Get k
  getK = const B.get


instance ( SerializableKnowledge k1, SerializableKnowledge k2 ) => SerializableKnowledge (k1, k2) where
  putK f (k1, k2) = putK f k1 >> putK f k2
  getK f = do
    k1 <- getK f
    k2 <- getK f
    return (k1, k2)

---------------------------------------- KnownData
-- TODO: change name
data KnownData = KnownData { 
    _kd_files :: !(SM.Map FileId FileData)
  , _kd_storages :: !StorageSeenMap
  , _kd_contents :: !(SM.Map StorageId SeenTree)
  , _kd_priorities :: !(SM.Map StorageId (PathNode (Maybe Priority)))
  } deriving Show

class KnownDataClass a where
  kd_files :: Lens' a (SM.Map FileId FileData)
  kd_storages :: Lens' a StorageSeenMap
  kd_contents :: Lens' a (SM.Map StorageId SeenTree)
  kd_priorities :: Lens' a (SM.Map StorageId (PathNode (Maybe Priority)))

instance KnownDataClass KnownData where
  kd_files = lens _kd_files $ \ s a -> s { _kd_files = a }
  kd_storages = lens _kd_storages $ \ s a -> s { _kd_storages = a }
  kd_contents = lens _kd_contents $ \ s a -> s { _kd_contents = a }
  kd_priorities = lens _kd_priorities $ \ s a -> s { _kd_priorities = a }

instance KnownDataClass (KnownData, b) where
  kd_files = _1 . kd_files
  kd_storages = _1 . kd_storages
  kd_contents = _1 . kd_contents
  kd_priorities = _1 . kd_priorities

instance KnownDataClass a => KnownDataClass (b, a) where
  kd_files = _2 . kd_files
  kd_storages = _2 . kd_storages
  kd_contents = _2 . kd_contents
  kd_priorities = _2 . kd_priorities

instance Knowledge KnownData where
  noKnowledge = KnownData SM.empty SM.empty SM.empty SM.empty
  applyFact = applyFact''
  kCode = const "k"

instance SerializableKnowledge KnownData where
  putK f ( KnownData a b c d ) = do
    B.put b
    B.put $ map ( mapFst $ fromJust' "Shorthand" . flip findShorthand f ) $ SM.toList a
    B.put $ toShared f c
    B.put d
    where
      toShared f = map ( mapSnd ( toShared' f ) ) . SM.toList 
      toShared' f = fmap ( fromJust' "Shorthand" . flip findShorthand f ) . toLazy
  getK f = do
    b <- getMap getStrict getStrict
    let sids = S.toList $ S.fromList $ SM.keys b
    let getSid sid = fromJust' "Deduplicate sid" $ find ( == sid ) sids
    let dedup fd = fd { fd_seen = S.fromList $ map ( mapFst getSid ) $ S.toList $ fd_seen fd }
    a <- getMap ( liftM ( force . head . flip findFid f ) B.get ) ( liftM dedup $ getStrict )
    c <- getMap getStrict $ ST.getSeenTree ( head . flip findFid f )
    d <- getMap getStrict getStrict
    return $ KnownData a b c d


applyFact'' :: Fact -> KnownData -> KnownData
applyFact'' ( FileSeen _ fid sid path ) = ( kd_contents %~ addToContent ) . ( kd_files %~ f )
  where
    f fd = SM.insert fid ( pre { fd_seen = S.insert ( sid, pathToShort path ) $ fd_seen pre } ) fd
      where
        pre = SM.findWithDefault noFileData fid fd
    addToContent = SM.insertWith ( const $ insertFile path fid ) sid ( insertFile path fid ST.empty )
applyFact'' ( FileMiss _ sid path ) = ( kd_contents %~ remFromContent )
  where
    -- TODO: also delete from kd_files
    remFromContent = SM.adjust ( removeFile path ) sid
applyFact'' ( FileAttr _ fid attrs ) = id

applyFact'' ( PathAttr _ sid path attrs ) = 
  case mapMaybe ( \ a -> case a of Require pri -> Just pri; _ -> Nothing ) attrs of
    [] -> id
    pri : _ -> execState $ do
      subTree <- use $ kd_contents . at sid . to ( fromMaybe ST.empty ) . to ( findNode $ splitDirectories path ) . to ( fromMaybe ST.empty )
      oldPris <- use $ kd_priorities . at sid . to ( fromMaybe PA.empty )
      let pathsFids = map ( mapFst $ joinPath . ( path : ) ) $ allFiles subTree
      flip mapM_ pathsFids $ \ ( path, fid ) -> do
        case oldPris ^. pathLens True path of
          Nothing ->
            kd_files . at fid %= Just . fd_withPriority pri . fromMaybe noFileData
          Just oldPri ->
            kd_files . at fid %= fmap ( fd_swapPriority oldPri pri )
      kd_priorities . at sid .= Just ( pathSet path .~ PathNode ( Just pri ) SM.empty $ oldPris )
applyFact'' ( StorageSeen _ts envid sid path ) = kd_storages %~ storageInsert envid sid path
  where
    storageInsert :: EnvironmentId -> StorageId -> FilePath -> StorageSeenMap -> StorageSeenMap
    storageInsert envid sid fp m =
      flip ( SM.insert sid ) m $ case SM.lookup sid m of
        Nothing -> [ (envid, fp) ]
        Just l -> ( envid, fp ) : filter ( ( envid, fp ) /= ) l
applyFact'' _ = id

instance B.Binary KnownData where
  put ( KnownData a b c d ) = B.put a >> B.put b >> B.put c >> B.put d
  get = liftM4 KnownData B.get B.get B.get B.get


---------------------------------------- FactCounter
data FactCounter = FactCounter { fc_c :: !Int }
                  deriving (Show, Eq)

class FactCounterClass a where
  fact_counter :: Lens' a Int

instance FactCounterClass FactCounter where
  fact_counter = lens fc_c $ const FactCounter

instance FactCounterClass ( FactCounter, b ) where
  fact_counter = _1 . fact_counter

instance FactCounterClass a => FactCounterClass ( b, a ) where
  fact_counter = _2 . fact_counter

instance Knowledge FactCounter where
  noKnowledge = FactCounter 0
  applyFact _ ( FactCounter i ) = FactCounter ( i + 1 )
  kCode = const "c"

instance B.Binary FactCounter where
  put = B.put . fc_c
  get = liftM FactCounter B.get

instance SerializableKnowledge FactCounter


---------------------------------------- FileSet
data FileSet = FileSet {
  _fs_files :: !( S.Set FileId ) 
  } deriving Show

class FileSetClass a where
  fs_files :: Lens' a ( S.Set FileId )

instance FileSetClass FileSet where
  fs_files = lens _fs_files $ const FileSet

instance FileSetClass ( FileSet, b ) where
  fs_files = _1 . fs_files

instance FileSetClass a => FileSetClass ( b, a ) where
  fs_files = _2 . fs_files

instance Knowledge FileSet where
  noKnowledge = FileSet S.empty
  applyFact f x @ ( FileSet s ) = case mfact_file f of
    Just fid -> FileSet $ S.insert fid s
    Nothing -> x
  kCode = const "f"

instance B.Binary FileSet where
  put = B.put . _fs_files
  get = liftM FileSet B.get

instance SerializableKnowledge FileSet


---------------------------------------- SeenStorages
data SeenStorages = SeenStorages {
  _ss_storages :: !( SM.Map StorageId [ String ] )
  } deriving Show

class SeenStoragesClass a where
  seen_storages :: Lens' a ( SM.Map StorageId [ String ] )

instance SeenStoragesClass SeenStorages where
  seen_storages = lens _ss_storages $ \ s a -> s { _ss_storages = a }

instance SeenStoragesClass ( SeenStorages, b ) where
  seen_storages = _1 . seen_storages

instance SeenStoragesClass a => SeenStoragesClass ( b, a ) where
  seen_storages = _2 . seen_storages

instance Knowledge SeenStorages where
  noKnowledge = SeenStorages SM.empty
  applyFact ( LabelStorage sid label ) ( SeenStorages s ) =
    SeenStorages ( SM.insertWith (++) sid [ label ] s )
  applyFact x ( SeenStorages s ) =
    case mfact_storage x of
      Just sid -> sid `deepseq` SeenStorages ( SM.insertWith (++) sid [] s )
      Nothing -> SeenStorages s
  kCode = const "s"

instance B.Binary SeenStorages where
  put = B.put . _ss_storages
  get = liftM SeenStorages B.get

instance SerializableKnowledge SeenStorages


---------------------------------------- SeenEnvironments
data SeenEnvironments = SeenEnvironments {
  _se_environments :: !( SM.Map StorageId [ String ] )
  } deriving Show

class SeenEnvironmentsClass a where
  seen_environments :: Lens' a ( SM.Map StorageId [ String ] )

instance SeenEnvironmentsClass SeenEnvironments where
  seen_environments = lens _se_environments $ \ s a -> s { _se_environments = a }

instance SeenEnvironmentsClass ( SeenEnvironments, b ) where
  seen_environments = _1 . seen_environments

instance SeenEnvironmentsClass a => SeenEnvironmentsClass ( b, a ) where
  seen_environments = _2 . seen_environments

instance Knowledge SeenEnvironments where
  noKnowledge = SeenEnvironments SM.empty
  applyFact ( LabelEnvironment envid label ) ( SeenEnvironments s ) =
    SeenEnvironments ( SM.insertWith (++) envid [ label ] s )
  applyFact x @ ( StorageSeen _ _ _ _ ) ( SeenEnvironments s ) =
    SeenEnvironments ( SM.insertWith (++) ( fact_env x ) [] s )
  applyFact _ s = s
  kCode = const "e"

instance B.Binary SeenEnvironments where
  put = B.put . _se_environments
  get = liftM SeenEnvironments B.get

instance SerializableKnowledge SeenEnvironments
